/**
@file
@copyright (c) 2023 EMBETECH SP. Z O.O. All rights reserved.
@version   1.1.4490
@purpose   embeNET demo
@brief     Exemplary task
*/

#include "blink_task.h"
#include "main.h"
#include "embenet_node.h"
#include <inttypes.h>
#include <stdio.h>
#include <string.h>

// How often the LED will blink (in ms)
#define BLINK_TASK_PERIOD_MS 		1500
// How long will the LED be on (in ms)
#define BLINK_TASK_ON_DURATION_MS 	200

// Task Id
static EMBENET_TaskId blinkOnTaskId = EMBENET_TASKID_INVALID;

// State of the LED (true = on, false = off)
static bool blinkLedIsOn;

/**
 * @brief User-defined function that will be invoked as a periodically scheduled task
 *
 * @param[in] taskId id of the task
 * @param[in] timeSource time source (local time or network time)
 * @param[in] t time at which the task was scheduled to run
 * @param[in] context generic, user-defined context
 */
static void blinkTask(EMBENET_TaskId taskId, EMBENET_NODE_TimeSource timeSource, uint64_t time, void *context) {
	(void) context;
	if (blinkLedIsOn) {
		// Turn off the LED
		HAL_GPIO_WritePin(GPIOB, LED3_Pin, GPIO_PIN_RESET);
		blinkLedIsOn = false;
		// Reschedule this task
		EMBENET_NODE_TaskSchedule(taskId, timeSource, time + BLINK_TASK_PERIOD_MS-BLINK_TASK_ON_DURATION_MS);
	} else {
		// Turn on the LED
		HAL_GPIO_WritePin(GPIOB, LED3_Pin, GPIO_PIN_SET);
		blinkLedIsOn = true;
		// Reschedule this task
		EMBENET_NODE_TaskSchedule(taskId, timeSource, time + BLINK_TASK_ON_DURATION_MS);
	}
}

void blink_task_init(void) {
	// creating LED blinking task
	if ( EMBENET_TASKID_INVALID == (blinkOnTaskId = EMBENET_NODE_TaskCreate(blinkTask, NULL))) {
		printf("Failed to create LED blink task");
	}
}

void blink_task_start(void) {
	// Get network time and calculate an even timestamp
	uint64_t timestamp = EMBENET_NODE_GetNetworkTime() / BLINK_TASK_PERIOD_MS + 1;
	timestamp = timestamp * BLINK_TASK_PERIOD_MS;
	// Schedule the blinking task
	if (EMBENET_RESULT_OK != EMBENET_NODE_TaskSchedule(blinkOnTaskId, EMBENET_NODE_TIME_SOURCE_NETWORK, timestamp)) {
		printf("Failed to start LED blink task");
	}
}

void blink_task_stop(void) {
	// Stop LED blinking task
	if (EMBENET_RESULT_OK != EMBENET_NODE_TaskCancel(blinkOnTaskId)) {
		printf("Failed to stop LED blink task");
	}
	// Return to off state
	HAL_GPIO_WritePin(GPIOB, LED3_Pin, GPIO_PIN_RESET);
	blinkLedIsOn = false;
}
